﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Text;

namespace Delay
{
    /// <summary>
    /// Class that implements the basic functionality of MsiExec.exe.
    /// </summary>
    internal class ManagedMsiExec
    {
        // Using statics for the two delegate types ensures they stay alive for the life of the application
        private static readonly Msi.InstallUiHandler _installUiHandler = InstallUiHandler;
        private static readonly Msi.InstallUiHandlerRecord _installUiHandlerRecord = InstallUiHandlerRecord;

        // Variables
        private static string _currentAction = "";
        private static int _maximumProgressLineLength;
        private static bool _suppressingProgressBarUpdates;
        private static int _expectedTicks;
        private static int _ticks;
        private static int _ticksPerActionData;

        /// <summary>
        /// Application entry point.
        /// </summary>
        /// <param name="args">Arguments.</param>
        /// <returns>Exit code.</returns>
        private static int Main(string[] args)
        {
            var returnValue = Msi.ERROR.INVALID_PARAMETER;
            bool installing;
            string packagePath;
            bool verbose;

            if (!ParseCommandLine(args, out installing, out packagePath, out verbose))
            {
                ConsoleWriteLine("SYNTAX: ManagedMsiExec <--Install|-i|--Uninstall|-u> Package.msi [--Verbose|-v]");
                returnValue = Msi.ERROR.INVALID_PARAMETER;
            }
            else
            {
                ConsoleWriteLine("ManagedMsiExec: {0}nstalling {1}", installing ? "I" : "Uni", packagePath);

                // Disable Windows Installer UI
                var windowHandle = IntPtr.Zero;
                var previousInstallUiLevel = Msi.MsiSetInternalUI(Msi.INSTALLUILEVEL.NONE, ref windowHandle);
                Debug.Assert(Msi.INSTALLUILEVEL.DEFAULT == previousInstallUiLevel);

                // Open the installer package
                var productHandle = IntPtr.Zero;
                try
                {
                    var openPackageResult = Msi.MsiOpenPackage(packagePath, out productHandle);
                    if (Msi.ERROR.SUCCESS != openPackageResult)
                    {
                        returnValue = openPackageResult;
                    }
                    else
                    {
                        if (verbose)
                        {
                            // Output all UI messages with default formatting
                            var previousExternalUIHandler = Msi.MsiSetExternalUI(
                                _installUiHandler,
                                Msi.INSTALLLOGMODE.FATALEXIT | Msi.INSTALLLOGMODE.ERROR | Msi.INSTALLLOGMODE.WARNING |
                                Msi.INSTALLLOGMODE.USER | Msi.INSTALLLOGMODE.INFO | Msi.INSTALLLOGMODE.RESOLVESOURCE |
                                Msi.INSTALLLOGMODE.OUTOFDISKSPACE | Msi.INSTALLLOGMODE.ACTIONSTART | Msi.INSTALLLOGMODE.ACTIONDATA |
                                Msi.INSTALLLOGMODE.COMMONDATA | Msi.INSTALLLOGMODE.PROGRESS | Msi.INSTALLLOGMODE.INITIALIZE |
                                Msi.INSTALLLOGMODE.TERMINATE | Msi.INSTALLLOGMODE.SHOWDIALOG | Msi.INSTALLLOGMODE.FILESINUSE |
                                Msi.INSTALLLOGMODE.RMFILESINUSE | Msi.INSTALLLOGMODE.INSTALLSTART | Msi.INSTALLLOGMODE.INSTALLEND,
                                IntPtr.Zero);
                            Debug.Assert(null == previousExternalUIHandler);
                        }
                        else
                        {
                            // Output select UI messages with custom formatting
                            Msi.InstallUiHandlerRecord previousExternalUiRecordHandler = null;
                            var setExternalUIRecordResult = Msi.MsiSetExternalUIRecord(
                                _installUiHandlerRecord,
                                Msi.INSTALLLOGMODE.FATALEXIT | Msi.INSTALLLOGMODE.ERROR | Msi.INSTALLLOGMODE.WARNING |
                                Msi.INSTALLLOGMODE.ACTIONSTART | Msi.INSTALLLOGMODE.ACTIONDATA | Msi.INSTALLLOGMODE.PROGRESS |
                                Msi.INSTALLLOGMODE.INSTALLEND,
                                productHandle,
                                out previousExternalUiRecordHandler);
                            Debug.Assert(Msi.ERROR.SUCCESS == setExternalUIRecordResult);
                        }

                        // Set CLIENTUILEVEL=0 to work around the "CLIENTUILEVEL= REMOVE=ALL" bug in Windows Installer
                        // http://stackoverflow.com/questions/9010145/duplicate-entries-in-add-removeprograms-control-panel-when-using-msisetexternalu
                        var setPropertyResult = Msi.MsiSetProperty(productHandle, "CLIENTUILEVEL", "0");
                        Debug.Assert(Msi.ERROR.SUCCESS == setPropertyResult);

                        if (!installing)
                        {
                            // Set REMOVE=ALL to perform an uninstall
                            setPropertyResult = Msi.MsiSetProperty(productHandle, "REMOVE", "ALL");
                            Debug.Assert(Msi.ERROR.SUCCESS == setPropertyResult);
                        }

                        // Perform the install/uninstall
                        var doActionResult = Msi.MsiDoAction(productHandle, "");
                        // Debug.Assert(Msi.ERROR.SUCCESS == doActionResult);
                        returnValue = doActionResult;
                    }
                }
                finally
                {
                    // Close the handle
                    var closeHandleResult = Msi.MsiCloseHandle(productHandle);
                    Debug.Assert(Msi.ERROR.SUCCESS == closeHandleResult);
                }
            }

            // Output result code/string
            ConsoleWriteLine("Windows Installer result: {0} ({1})", (int)returnValue, returnValue);
            return (int)returnValue;
        }

        /// <summary>
        /// Handles install UI messages by displaying them directly.
        /// </summary>
        /// <param name="pvContext">Context.</param>
        /// <param name="iMessageType">Message type.</param>
        /// <param name="szMessage">Message.</param>
        /// <returns>0 to indicate no action taken.</returns>
        private static int InstallUiHandler(IntPtr pvContext, uint iMessageType, string szMessage)
        {
            ConsoleWriteLine("{0}: {1}", InstallMessageValueToString(iMessageType), szMessage);
            return 0;
        }

        /// <summary>
        /// Handles install UI messages by formatting and displaying them.
        /// </summary>
        /// <param name="pvContext">Context.</param>
        /// <param name="iMessageType">Message type.</param>
        /// <param name="hRecord">Record handle.</param>
        /// <returns>0 to indicate no action taken.</returns>
        private static int InstallUiHandlerRecord(IntPtr pvContext, uint iMessageType, IntPtr hRecord)
        {
            // Initialize from parameters
            var productHandle = pvContext;
            var installMessage = (Msi.INSTALLMESSAGE)(Msi.INSTALLMESSAGE_MASK & iMessageType);

            // Local variables
            var skipOutput = false;
            var showProgressBar = true;
            var buffer = new StringBuilder();
            var bufferLength = (uint)buffer.Capacity;

            if (Msi.INSTALLMESSAGE.ACTIONSTART == installMessage)
            {
                // Extract action name
                var recordGetStringResult = Msi.MsiRecordGetString(hRecord, 1, buffer, ref bufferLength);
                if (Msi.ERROR.MORE_DATA == recordGetStringResult)
                {
                    bufferLength++; // Return value does not include trailing NULL
                    buffer.Length = (int)bufferLength;
                    recordGetStringResult = Msi.MsiRecordGetString(hRecord, 1, buffer, ref bufferLength);
                }
                if (Msi.ERROR.SUCCESS == recordGetStringResult)
                {
                    _currentAction = buffer.ToString();
                }
            }
            else if (Msi.INSTALLMESSAGE.ACTIONDATA == installMessage)
            {
                // Increment ticks for explicit progress mode
                _ticks += _ticksPerActionData;
            }
            else if (Msi.INSTALLMESSAGE.PROGRESS == installMessage)
            {
                // Update progress bar
                var fieldCount = Msi.MsiRecordGetFieldCount(hRecord);
                if ((0 == fieldCount) || (uint.MaxValue == fieldCount))
                {
                    skipOutput = true;
                }
                else
                {
                    // Read fields
                    var field1 = Msi.MsiRecordGetInteger(hRecord, 1);
                    Debug.Assert(Msi.MSI_NULL_INTEGER != field1);
                    var field2 = Msi.MsiRecordGetInteger(hRecord, 2);
                    Debug.Assert(Msi.MSI_NULL_INTEGER != field2);
                    var field3 = Msi.MsiRecordGetInteger(hRecord, 3);
                    Debug.Assert((fieldCount < 3) || (Msi.MSI_NULL_INTEGER != field3));
                    var field4 = Msi.MsiRecordGetInteger(hRecord, 4);
                    Debug.Assert((fieldCount < 4) || (Msi.MSI_NULL_INTEGER != field4));

                    // Parse progress data
                    if (0 == field1)
                    {
                        _expectedTicks = field2;
                        // var backward = 1 == field3;
                        // var inProgress = 0 == field4;
                        _ticks = 0;
                    }
                    else if (1 == field1)
                    {
                        var explicitProgress = 0 == field3;
                        if (!explicitProgress)
                        {
                            _ticksPerActionData = field2;
                        }
                    }
                    else if (2 == field1)
                    {
                        var addTicks = field2;
                        _ticks += addTicks;
                    }
                    else if (3 == field1)
                    {
                        var additionalTicks = field1;
                        _expectedTicks += additionalTicks;
                    }
                    else
                    {
                        // Unexpected input
                        Debug.Assert(false);
                    }
                }
            }
            else if (Msi.INSTALLMESSAGE.INSTALLEND == installMessage)
            {
                // Complete output
                ConsoleWrite("\r{0}\r", "".PadRight(_maximumProgressLineLength));
                skipOutput = true;
            }
            else
            {
                // Failure/warning message
                Debug.Assert((Msi.INSTALLMESSAGE.ERROR == installMessage) || (Msi.INSTALLMESSAGE.WARNING == installMessage) || (Msi.INSTALLMESSAGE.FATALEXIT == installMessage));
                showProgressBar = false;
                _suppressingProgressBarUpdates = true;

                // Extract message
                var formatRecordResult = Msi.MsiFormatRecord(productHandle, hRecord, buffer, ref bufferLength);
                if (Msi.ERROR.MORE_DATA == formatRecordResult)
                {
                    bufferLength++; // Return value does not include trailing NULL
                    buffer.Length = (int)bufferLength;
                    formatRecordResult = Msi.MsiFormatRecord(productHandle, hRecord, buffer, ref bufferLength);
                }
                if (Msi.ERROR.SUCCESS == formatRecordResult)
                {
                    _currentAction = buffer.ToString();
                }
            }

            // Display status
            if (!skipOutput)
            {
                if (showProgressBar)
                {
                    if (!_suppressingProgressBarUpdates)
                    {
                        var progressBarLengthTotal = 20;
                        var progressBarLength = Math.Min((0 < _expectedTicks) ? (progressBarLengthTotal * _ticks) / _expectedTicks : 0, progressBarLengthTotal);
                        var progressBar = "".PadRight(progressBarLength, '#').PadRight(progressBarLengthTotal, ' ');
                        var progressLine = string.Format(CultureInfo.CurrentCulture, "[{0}] {1}", progressBar, _currentAction);
                        ConsoleWrite("\r{0}", progressLine.PadRight(_maximumProgressLineLength));
                        _maximumProgressLineLength = Math.Max(progressLine.Length, _maximumProgressLineLength);
                    }
                }
                else
                {
                    ConsoleWriteLine("\n{0}", _currentAction);
                }
            }

            // No action taken
            return 0;
        }

        /// <summary>
        /// Parse an install message value and return its string representation.
        /// </summary>
        /// <param name="installMessage">Message type.</param>
        /// <returns>String representation.</returns>
        private static string InstallMessageValueToString(uint installMessage)
        {
            installMessage &= Msi.INSTALLMESSAGE_MASK;
            foreach (var value in Enum.GetValues(typeof(Msi.INSTALLMESSAGE)))
            {
                if ((int)value == installMessage)
                {
                    return Enum.GetName(typeof(Msi.INSTALLMESSAGE), value);
                }
            }
            return "???";
        }

        /// <summary>
        /// Parse command line arguments.
        /// </summary>
        /// <param name="args">Arguments.</param>
        /// <param name="installing">True iff installing.</param>
        /// <param name="packagePath">Package path.</param>
        /// <param name="verbose">True iff verbose.</param>
        /// <returns>True iff valid syntax.</returns>
        private static bool ParseCommandLine(IEnumerable<string> args, out bool installing, out string packagePath, out bool verbose)
        {
            // Initialize
            installing = true;
            packagePath = null;
            verbose = false;
            var argsEnumerator = args.GetEnumerator();

            // Parse install/uninstall switch
            if (!argsEnumerator.MoveNext())
            {
                return false;
            }
            var installUninstallString = argsEnumerator.Current;
            if ((string.Equals("--install", installUninstallString, StringComparison.OrdinalIgnoreCase)) ||
                 string.Equals("-i", installUninstallString, StringComparison.OrdinalIgnoreCase))
            {
                installing = true;
            }
            else if ((string.Equals("--uninstall", installUninstallString, StringComparison.OrdinalIgnoreCase)) ||
                      string.Equals("-u", installUninstallString, StringComparison.OrdinalIgnoreCase))
            {
                installing = false;
            }
            else
            {
                return false;
            }

            // Parse package path
            if (!argsEnumerator.MoveNext())
            {
                return false;
            }
            packagePath = argsEnumerator.Current;
            if (string.IsNullOrWhiteSpace(packagePath))
            {
                return false;
            }

            // Get full path to avoid MSI error 1619
            packagePath = Path.GetFullPath(packagePath);

            // Parse verbose switch
            if (argsEnumerator.MoveNext())
            {
                var verboseString = argsEnumerator.Current;
                if ((string.Equals("--verbose", verboseString, StringComparison.OrdinalIgnoreCase)) ||
                     string.Equals("-v", verboseString, StringComparison.OrdinalIgnoreCase))
                {
                    verbose = true;
                }
                else
                {
                    return false;
                }
                if (argsEnumerator.MoveNext())
                {
                    return false;
                }
            }

            // Return success
            return true;
        }

        /// <summary>
        /// Writes a formatted string to the console.
        /// </summary>
        /// <param name="format">Format string.</param>
        /// <param name="arg">Arguments.</param>
        private static void ConsoleWrite(string format, params object[] arg)
        {
            Console.Write(format, arg);
        }

        /// <summary>
        /// Writes a formatted line to the console.
        /// </summary>
        /// <param name="format">Format string.</param>
        /// <param name="arg">Arguments.</param>
        private static void ConsoleWriteLine(string format, params object[] arg)
        {
            Console.WriteLine(format, arg);
        }
    }
}
